import importlib
import inspect
import os
from abc import ABC
from collections import OrderedDict
from collections.abc import Callable, Iterable, Iterator, Sequence
from enum import Enum
from functools import partial
from typing import Any, Self

import numpy as np
import torch
from torch._jit_internal import _copy_to_script_wrapper
from torch.utils.checkpoint import checkpoint

from konfai import konfai_root
from konfai.data.patching import Accumulator, ModelPatch
from konfai.metric.schedulers import Scheduler
from konfai.utils.config import config
from konfai.utils.utils import MeasureError, State, TrainerError, get_device, get_gpu_memory, get_module


class NetState(Enum):
    TRAIN = (0,)
    PREDICTION = 1


class PatchIndexed:

    def __init__(self, patch: ModelPatch, index: int) -> None:
        self.patch = patch
        self.index = index

    def is_full(self) -> bool:
        return len(self.patch.get_patch_slices(0)) == self.index


class OptimizerLoader:

    @config("Optimizer")
    def __init__(self, name: str = "AdamW") -> None:
        self.name = name

    def get_optimizer(self, key: str, parameter: Iterator[torch.nn.parameter.Parameter]) -> torch.optim.Optimizer:
        return config(f"{konfai_root()}.Model.{key}.Optimizer")(
            getattr(importlib.import_module("torch.optim"), self.name)
        )(parameter, config=None)


class LRSchedulersLoader:

    @config()
    def __init__(self, nb_step: int = 0) -> None:
        self.nb_step = nb_step

    def getschedulers(
        self, key: str, scheduler_classname: str, optimizer: torch.optim.Optimizer
    ) -> torch.optim.lr_scheduler._LRScheduler:
        for m in ["torch.optim.lr_scheduler", "konfai.metric.schedulers"]:
            module, name = get_module(scheduler_classname, m)
            if hasattr(importlib.import_module(module), name):
                return config(f"{konfai_root()}.Model.{key}.schedulers.{scheduler_classname}")(
                    getattr(importlib.import_module(module), name)
                )(optimizer, config=None)
        raise TrainerError(
            f"Unknown scheduler {scheduler_classname}, tried importing from: 'torch.optim.lr_scheduler' and "
            "'konfai.metric.schedulers', but no valid match was found. "
            "Check your YAML config or scheduler name spelling."
        )


class LossSchedulersLoader:

    @config()
    def __init__(self, nb_step: int = 0) -> None:
        self.nb_step = nb_step

    def getschedulers(self, key: str, scheduler_classname: str) -> torch.optim.lr_scheduler._LRScheduler:
        return config(f"{key}.{scheduler_classname}")(
            getattr(importlib.import_module("konfai.metric.schedulers"), scheduler_classname)
        )(config=None)


class CriterionsAttr:

    @config()
    def __init__(
        self,
        schedulers: dict[str, LossSchedulersLoader] = {"default:Constant": LossSchedulersLoader(0)},
        is_loss: bool = True,
        group: int = 0,
        start: int = 0,
        stop: int | None = None,
        accumulation: bool = False,
    ) -> None:
        self.schedulersLoader = schedulers
        self.isTorchCriterion = True
        self.is_loss = is_loss
        self.start = start
        self.stop = stop
        self.group = group
        self.accumulation = accumulation
        self.schedulers: dict[Scheduler, int] = {}


class CriterionsLoader:

    @config()
    def __init__(
        self,
        criterions_loader: dict[str, CriterionsAttr] = {"default:torch:nn:CrossEntropyLoss:Dice:NCC": CriterionsAttr()},
    ) -> None:
        self.criterions_loader = criterions_loader

    def get_criterions(
        self, model_classname: str, output_group: str, target_group: str
    ) -> dict[torch.nn.Module, CriterionsAttr]:
        criterions = {}
        for module_classpath, criterions_attr in self.criterions_loader.items():
            module, name = get_module(module_classpath, "konfai.metric.measure")
            criterions_attr.isTorchCriterion = module.startswith("torch")
            for (
                scheduler_classname,
                schedulers,
            ) in criterions_attr.schedulersLoader.items():
                criterions_attr.schedulers[
                    schedulers.getschedulers(
                        f"{konfai_root()}.Model.{model_classname}.outputs_criterions.{output_group}"
                        f".targets_criterions.{target_group}"
                        f".criterions_loader.{module_classpath}.schedulers",
                        scheduler_classname,
                    )
                ] = schedulers.nb_step
            criterions[
                config(
                    f"{konfai_root()}.Model.{model_classname}.outputs_criterions."
                    f"{output_group}.targets_criterions.{target_group}."
                    f"criterions_loader.{module_classpath}"
                )(getattr(importlib.import_module(module), name))(config=None)
            ] = criterions_attr
        return criterions


class TargetCriterionsLoader:

    @config()
    def __init__(
        self,
        targets_criterions: dict[str, CriterionsLoader] = {"default": CriterionsLoader()},
    ) -> None:
        self.targets_criterions = targets_criterions

    def get_targets_criterions(
        self, output_group: str, model_classname: str
    ) -> dict[str, dict[torch.nn.Module, CriterionsAttr]]:
        targets_criterions = {}
        for target_group, criterions_loader in self.targets_criterions.items():
            targets_criterions[target_group] = criterions_loader.get_criterions(
                model_classname, output_group, target_group
            )
        return targets_criterions


class Measure:

    class Loss:

        def __init__(
            self,
            name: str,
            output_group: str,
            target_group: str,
            group: int,
            is_loss: bool,
            accumulation: bool,
        ) -> None:
            self.name = name
            self.is_loss = is_loss
            self.accumulation = accumulation
            self.output_group = output_group
            self.target_group = target_group
            self.group = group

            self._loss: list[torch.Tensor] = []
            self._weight: list[float] = []
            self._values: list[float] = []

        def reset_loss(self) -> None:
            self._loss.clear()

        def add(self, weight: float, value: torch.Tensor | tuple[torch.Tensor, float]) -> None:
            if isinstance(value, tuple):
                loss_value, true_value = value
            else:
                loss_value = value
                true_value = value.item()

            self._loss.append(loss_value if self.is_loss else loss_value.detach())
            self._values.append(true_value)
            self._weight.append(weight)

        def get_last_loss(self) -> torch.Tensor:
            return self._loss[-1] * self._weight[-1] if len(self._loss) else torch.zeros((1), requires_grad=True)

        def get_loss(self) -> torch.Tensor:
            return (
                torch.stack([w * loss_value for w, loss_value in zip(self._weight, self._loss)], dim=0).mean(dim=0)
                if len(self._loss)
                else torch.zeros((1), requires_grad=True)
            )

        def __len__(self) -> int:
            return len(self._loss)

    def __init__(
        self,
        model_classname: str,
        outputs_criterions_loader: dict[str, TargetCriterionsLoader],
    ) -> None:
        super().__init__()
        self.outputs_criterions: dict[str, dict[str, dict[torch.nn.Module, CriterionsAttr]]] = {}
        for output_group, target_criterions_loader in outputs_criterions_loader.items():
            self.outputs_criterions[output_group.replace(":", ".")] = target_criterions_loader.get_targets_criterions(
                output_group, model_classname
            )
        self._loss: dict[int, dict[str, Measure.Loss]] = {}

    def init(self, model: torch.nn.Module, group_dest: list[str]) -> None:
        outputs_group_rename = {}

        modules = []
        for i, _, _ in model.named_module_args_dict():
            modules.append(i)

        for output_group in self.outputs_criterions.keys():
            if output_group.replace(";accu;", "") not in modules:
                raise MeasureError(
                    f"The output group '{output_group}' defined in 'outputs_criterions' "
                    "does not correspond to any module in the model.",
                    f"Available modules: {modules}",
                    "Please check that the name matches exactly a submodule or output of your model architecture.",
                )
            for target_group in self.outputs_criterions[output_group]:
                for target_group_tmp in target_group.split(";"):
                    if target_group_tmp not in group_dest:
                        raise MeasureError(
                            f"The target_group {target_group_tmp} defined in "
                            "'outputs_criterions.{output_group}.targets_criterions'"
                            " was not found in the available destination groups.",
                            "This target_group is expected for loss or metric computation, "
                            "but was not loaded in 'group_dest'.",
                            f"Please make sure that the group {target_group_tmp} is defined in "
                            "Dataset:groups_src:...:groups_dest: {target_group_tmp} "
                            "and correctly loaded from the dataset.",
                        )
                for criterion in self.outputs_criterions[output_group][target_group]:
                    if not self.outputs_criterions[output_group][target_group][criterion].isTorchCriterion:
                        outputs_group_rename[output_group] = criterion.init(model, output_group, target_group)

        outputs_criterions_bak = self.outputs_criterions.copy()
        for old, new in outputs_group_rename.items():
            self.outputs_criterions.pop(old)
            self.outputs_criterions[new] = outputs_criterions_bak[old]
        for output_group in self.outputs_criterions:
            for target_group in self.outputs_criterions[output_group]:
                for criterion, criterions_attr in self.outputs_criterions[output_group][target_group].items():
                    if criterions_attr.group not in self._loss:
                        self._loss[criterions_attr.group] = {}
                    self._loss[criterions_attr.group][
                        f"{output_group}:{target_group}:{criterion.__class__.__name__}"
                    ] = Measure.Loss(
                        criterion.__class__.__name__,
                        output_group,
                        target_group,
                        criterions_attr.group,
                        criterions_attr.is_loss,
                        criterions_attr.accumulation,
                    )

    def update(
        self,
        output_group: str,
        output: torch.Tensor,
        data_dict: dict[str, torch.Tensor],
        it: int,
        nb_patch: int,
        training: bool,
    ) -> None:
        for target_group in self.outputs_criterions[output_group]:
            target = [
                data_dict[group].to(output[0].device).detach()
                for group in target_group.split(";")
                if group in data_dict
            ]

            for criterion, criterions_attr in self.outputs_criterions[output_group][target_group].items():
                if it >= criterions_attr.start and (criterions_attr.stop is None or it <= criterions_attr.stop):
                    scheduler = self.update_scheduler(criterions_attr.schedulers, it)
                    self._loss[criterions_attr.group][
                        f"{output_group}:{target_group}:{criterion.__class__.__name__}"
                    ].add(scheduler.get_value(), criterion(output, *target))
                    if (
                        training
                        and len(
                            np.unique(
                                [
                                    len(loss_value)
                                    for loss_value in self._loss[criterions_attr.group].values()
                                    if loss_value.accumulation and loss_value.is_loss
                                ]
                            )
                        )
                        == 1
                    ):
                        if criterions_attr.is_loss:
                            loss = torch.zeros((1), requires_grad=True)
                            for v in [
                                loss_value
                                for loss_value in self._loss[criterions_attr.group].values()
                                if loss_value.accumulation and loss_value.is_loss
                            ]:
                                loss_value = v.get_last_loss()
                                loss = loss.to(loss_value.device) + loss_value
                            loss = loss / nb_patch
                            loss.backward()

    def get_loss(self) -> list[torch.Tensor]:
        loss: dict[int, torch.Tensor] = {}
        for group in self._loss.keys():
            loss[group] = torch.zeros((1), requires_grad=True)
            for v in self._loss[group].values():
                if v.is_loss and not v.accumulation:
                    loss_value = v.get_loss()
                    loss[v.group] = loss[v.group].to(loss_value.device) + loss_value
        return list(loss.values())

    def reset_loss(self) -> None:
        for group in self._loss.keys():
            for v in self._loss[group].values():
                v.reset_loss()

    def get_last_values(self, n: int = 1) -> dict[str, float]:
        result = {}
        for group in self._loss.keys():
            result.update(
                {
                    name: np.nanmean(value._values[-n:] if n > 0 else value._values)
                    for name, value in self._loss[group].items()
                    if n < 0 or len(value._values) >= n
                }
            )
        return result

    def get_last_weights(self, n: int = 1) -> dict[str, float]:
        result = {}
        for group in self._loss.keys():
            result.update(
                {
                    name: np.nanmean(value._weight[-n:] if n > 0 else value._weight)
                    for name, value in self._loss[group].items()
                    if n < 0 or len(value._values) >= n
                }
            )
        return result

    def format_loss(self, is_loss: bool, n: int) -> dict[str, tuple[float, float]]:
        result = {}
        for group in self._loss.keys():
            for name, loss in self._loss[group].items():
                if loss.is_loss == is_loss and len(loss._values) >= n:
                    result[name] = (
                        np.nanmean(loss._weight[-n:]),
                        np.nanmean(loss._values[-n:]),
                    )
        return result

    def update_scheduler(self, schedulers: dict[Scheduler, int], it: int) -> Scheduler:
        step = 0
        _scheduler = None
        for _scheduler, value in schedulers.items():
            if value is None or (it >= step and it < step + value):
                break
            step += value
        if _scheduler:
            _scheduler.step(it - step)
        if _scheduler is None:
            raise NameError(
                f"No scheduler found for iteration {it}. "
                f"Available steps were: {list(schedulers.values())}. "
                f"Check your configuration."
            )
        return _scheduler


class ModuleArgsDict(torch.nn.Module, ABC):

    class ModuleArgs:

        def __init__(
            self,
            in_branch: list[str],
            out_branch: list[str],
            pretrained: bool,
            alias: list[str],
            requires_grad: bool | None,
            training: None | bool,
        ) -> None:
            super().__init__()
            self.alias = alias
            self.pretrained = pretrained
            self.in_branch = in_branch
            self.out_branch = out_branch
            self.in_channels: int | None = None
            self.in_is_channel: bool = True
            self.out_channels: int | None = None
            self.out_is_channel: bool = True
            self.requires_grad = requires_grad
            self.isCheckpoint = False
            self.isGPU_Checkpoint = False
            self.gpu = "cpu"
            self.training = training
            self._isEnd = False

    def __init__(self) -> None:
        super().__init__()
        self._modulesArgs: dict[str, ModuleArgsDict.ModuleArgs] = {}
        self._training = NetState.TRAIN

    def _addindent(self, s_: str, num_spaces: int):
        s = s_.split("\n")
        if len(s) == 1:
            return s_
        first = s.pop(0)
        s = [(num_spaces * " ") + line for line in s]
        return first + "\n" + "\n".join(s)

    def __repr__(self):
        extra_lines = []

        extra_repr = self.extra_repr()
        if extra_repr:
            extra_lines = extra_repr.split("\n")

        child_lines = []

        def is_simple_branch(x):
            return len(x) > 1 or x[0] != 0

        for key, module in self._modules.items():
            mod_str = repr(module)

            mod_str = self._addindent(mod_str, 2)
            desc = ""
            if is_simple_branch(self._modulesArgs[key].in_branch) or is_simple_branch(
                self._modulesArgs[key].out_branch
            ):
                desc += f", {self._modulesArgs[key].in_branch}->{self._modulesArgs[key].out_branch}"
            if not self._modulesArgs[key].pretrained:
                desc += ", pretrained=False"
            if self._modulesArgs[key].alias:
                desc += f", alias={self._modulesArgs[key].alias}"
            desc += f", in_channels={self._modulesArgs[key].in_channels}"
            desc += f", in_is_channel={self._modulesArgs[key].in_is_channel}"
            desc += f", out_channels={self._modulesArgs[key].out_channels}"
            desc += f", out_is_channel={self._modulesArgs[key].out_is_channel}"
            desc += f", is_end={self._modulesArgs[key]._isEnd}"
            desc += f", isInCheckpoint={self._modulesArgs[key].isCheckpoint}"
            desc += f", isInGPU_Checkpoint={self._modulesArgs[key].isGPU_Checkpoint}"
            desc += f", requires_grad={self._modulesArgs[key].requires_grad}"
            desc += f", device={self._modulesArgs[key].gpu}"

            child_lines.append(f"({key}{desc}) {mod_str}")

        lines = extra_lines + child_lines

        desc = ""
        if lines:
            if len(extra_lines) == 1 and not child_lines:
                desc += extra_lines[0]
            else:
                desc += "\n  " + "\n  ".join(lines) + "\n"

        return f"{self._get_name()}({desc})"

    def __getitem__(self, key: str) -> torch.nn.Module:
        module = self._modules[key]
        if not module:
            raise ValueError(f"Module '{key}' is None or missing in self._modules")
        return module

    @_copy_to_script_wrapper
    def keys(self) -> Iterable[str]:
        return self._modules.keys()

    @_copy_to_script_wrapper
    def items(self) -> Iterable[tuple[str, torch.nn.Module | None]]:
        return self._modules.items()

    @_copy_to_script_wrapper
    def values(self) -> Iterable[torch.nn.Module | None]:
        return self._modules.values()

    def add_module(
        self,
        name: str,
        module: torch.nn.Module,
        in_branch: Sequence[int | str] = [0],
        out_branch: Sequence[int | str] = [0],
        pretrained: bool = True,
        alias: list[str] = [],
        requires_grad: bool | None = None,
        training: None | bool = None,
    ) -> None:
        super().add_module(name, module)
        self._modulesArgs[name] = ModuleArgsDict.ModuleArgs(
            [str(value) for value in in_branch],
            [str(value) for value in out_branch],
            pretrained,
            alias,
            requires_grad,
            training,
        )

    def get_mapping(self):
        results: dict[str, str] = {}
        for name, module_args in self._modulesArgs.items():
            module = self[name]
            if isinstance(module, ModuleArgsDict):
                if len(module_args.alias):
                    count = dict.fromkeys(set(module.get_mapping().values()), 0)
                    if len(count):
                        for k, v in module.get_mapping().items():
                            alias_name = module_args.alias[count[v]]
                            if k == "":
                                results.update({alias_name: name + "." + v})
                            else:
                                results.update({alias_name + "." + k: name + "." + v})
                            count[v] += 1
                    else:
                        for alias in module_args.alias:
                            results.update({alias: name})
                else:
                    results.update({k: name + "." + v for k, v in module.get_mapping().items()})
            else:
                for alias in module_args.alias:
                    results[alias] = name
        return results

    @staticmethod
    def init_func(module: torch.nn.Module, init_type: str, init_gain: float):
        if not isinstance(module, Network):
            if isinstance(module, ModuleArgsDict):
                module.init(init_type, init_gain)
            elif isinstance(module, torch.nn.modules.conv._ConvNd) or isinstance(module, torch.nn.Linear):
                if init_type == "normal":
                    torch.nn.init.normal_(module.weight, 0.0, init_gain)
                elif init_type == "xavier":
                    torch.nn.init.xavier_normal_(module.weight, gain=init_gain)
                elif init_type == "kaiming":
                    torch.nn.init.kaiming_normal_(module.weight, a=0, mode="fan_in")
                elif init_type == "orthogonal":
                    torch.nn.init.orthogonal_(module.weight, gain=init_gain)
                elif init_type == "trunc_normal":
                    torch.nn.init.trunc_normal_(module.weight, std=init_gain)
                else:
                    raise NotImplementedError(f"Initialization method {init_type} is not implemented")
                if module.bias is not None:
                    torch.nn.init.constant_(module.bias, 0.0)

            elif isinstance(module, torch.nn.modules.batchnorm._BatchNorm):
                if module.weight is not None:
                    torch.nn.init.normal_(module.weight, 0.0, std=init_gain)
                if module.bias is not None:
                    torch.nn.init.constant_(module.bias, 0.0)

    def init(self, init_type: str, init_gain: float):
        for module in self._modules.values():
            ModuleArgsDict.init_func(module, init_type, init_gain)

    def named_forward(self, *inputs: torch.Tensor) -> Iterator[tuple[str, torch.Tensor]]:
        if len(inputs) > 0:
            branchs: dict[str, torch.Tensor] = {}
            for i, sinput in enumerate(inputs):
                branchs[str(i)] = sinput

            out = inputs[0]
            tmp = []
            for name, module in self.items():
                if self._modulesArgs[name].training is None or (
                    not (self._modulesArgs[name].training and self._training == NetState.PREDICTION)
                    and not (not self._modulesArgs[name].training and self._training == NetState.TRAIN)
                ):
                    requires_grad = self._modulesArgs[name].requires_grad
                    if requires_grad is not None and module:
                        module.requires_grad_(requires_grad)
                    for ib in self._modulesArgs[name].in_branch:
                        if ib not in branchs:
                            branchs[ib] = inputs[0]
                    for branchs_key in branchs.keys():
                        if (
                            str(self._modulesArgs[name].gpu) != "cpu"
                            and str(branchs[branchs_key].device) != "cuda:" + self._modulesArgs[name].gpu
                        ):
                            branchs[branchs_key] = branchs[branchs_key].to(int(self._modulesArgs[name].gpu))

                    if self._modulesArgs[name].isCheckpoint:
                        out = checkpoint(
                            module,
                            *[branchs[i] for i in self._modulesArgs[name].in_branch],
                            use_reentrant=True,
                        )
                        for ob in self._modulesArgs[name].out_branch:
                            branchs[ob] = out
                        yield name, out
                    else:
                        if isinstance(module, ModuleArgsDict):
                            for k, out in module.named_forward(
                                *[branchs[i] for i in self._modulesArgs[name].in_branch]
                            ):
                                for ob in self._modulesArgs[name].out_branch:
                                    if ob in module._modulesArgs[k.split(".")[0].replace(";accu;", "")].out_branch:
                                        tmp.append(ob)
                                        branchs[ob] = out
                                yield name + "." + k, out
                            for ob in self._modulesArgs[name].out_branch:
                                if ob not in tmp:
                                    branchs[ob] = out
                        elif isinstance(module, torch.nn.Module):
                            out = module(*[branchs[i] for i in self._modulesArgs[name].in_branch])
                            for ob in self._modulesArgs[name].out_branch:
                                branchs[ob] = out
                            yield name, out
            del branchs

    def forward(self, *input: torch.Tensor) -> torch.Tensor:
        _v = input
        for _, _v in self.named_forward(*input):
            pass
        return _v

    def named_parameters(
        self, pretrained: bool = False, recurse=False
    ) -> Iterator[tuple[str, torch.nn.parameter.Parameter]]:
        for name, module_args in self._modulesArgs.items():
            module = self[name]
            if isinstance(module, ModuleArgsDict):
                for k, v in module.named_parameters(pretrained=pretrained):
                    yield name + "." + k, v
            elif isinstance(module, torch.nn.Module):
                if not pretrained or not module_args.pretrained:
                    if module_args.training is None or module_args.training:
                        for k, v in module.named_parameters():
                            yield name + "." + k, v

    def parameters(self, pretrained: bool = False):
        for _, v in self.named_parameters(pretrained=pretrained):
            yield v

    def named_module_args_dict(self) -> Iterator[tuple[str, Self, ModuleArgs]]:
        for name, module in self._modules.items():
            yield name, module, self._modulesArgs[name]
            if isinstance(module, ModuleArgsDict):
                for k, v, u in module.named_module_args_dict():
                    yield name + "." + k, v, u

    def _requires_grad(self, keys: list[str]):
        keys = keys.copy()
        for name, module, args in self.named_module_args_dict():
            requires_grad = args.requires_grad
            if requires_grad is not None:
                module.requires_grad_(requires_grad)
            if name in keys:
                keys.remove(name)
                if len(keys) == 0:
                    break


class OutputsGroup(list):

    def __init__(self, measure: Measure) -> None:
        self.layers: dict[str, torch.Tensor] = {}
        self.measure = measure

    def add_layer(self, name: str, layer: torch.Tensor):
        self.layers[name] = layer

    def is_done(self):
        return len(self) == len(self.layers)

    def clear(self):
        self.layers.clear()


class Network(ModuleArgsDict, ABC):

    def _apply_network(
        self,
        name_function: Callable[[Self], str],
        networks: list[str],
        key: str,
        function: Callable,
        *args,
        **kwargs,
    ) -> dict[str, object]:
        results: dict[str, object] = {}
        for module in self.values():
            if isinstance(module, Network):
                if name_function(module) not in networks:
                    networks.append(name_function(module))
                    for k, v in module._apply_network(
                        name_function,
                        networks,
                        key + "." + name_function(module),
                        function,
                        *args,
                        **kwargs,
                    ).items():
                        results.update({name_function(self) + "." + k: v})
        if len([param.name for param in list(inspect.signature(function).parameters.values()) if param.name == "key"]):
            function = partial(function, key=key)

        results[name_function(self)] = function(self, *args, **kwargs)
        return results

    def _function_network():  # type: ignore[misc]
        def _function_network_d(function: Callable):
            def new_function(self: Self, *args, **kwargs) -> dict[str, object]:
                return self._apply_network(
                    lambda network: network.get_name(),
                    [],
                    self.get_name(),
                    function,
                    *args,
                    **kwargs,
                )

            return new_function

        return _function_network_d

    def __init__(
        self,
        in_channels: int = 1,
        optimizer: OptimizerLoader | None = None,
        schedulers: dict[str, LRSchedulersLoader] | None = None,
        outputs_criterions: dict[str, TargetCriterionsLoader] | None = None,
        patch: ModelPatch | None = None,
        nb_batch_per_step: int = 1,
        init_type: str = "normal",
        init_gain: float = 0.02,
        dim: int = 3,
    ) -> None:
        super().__init__()
        self.name = self.__class__.__name__
        self.in_channels = in_channels
        self.optimizerLoader = optimizer
        self.optimizer: torch.optim.Optimizer | None = None

        self.lr_schedulers_loader = schedulers
        self.schedulers: dict[torch.optim.lr_scheduler._LRScheduler, int] = {}

        self.outputs_criterions_loader = outputs_criterions
        self.measure: Measure | None = None

        self.patch = patch

        self.nb_batch_per_step = nb_batch_per_step
        self.init_type = init_type
        self.init_gain = init_gain
        self.dim = dim
        self._it = 0
        self._nb_lr_update = 0
        self.outputsGroup: list[OutputsGroup] = []

    @_function_network()
    def state_dict(self) -> dict[str, OrderedDict]:
        destination: OrderedDict[str, Any] = OrderedDict()
        local_metadata = {"version": self._version}
        # destination["_metadata"] = OrderedDict({"": local_metadata})
        self._save_to_state_dict(destination, "", False)
        for name, module in self._modules.items():
            if module is not None:
                if not isinstance(module, Network):
                    module.state_dict(destination=destination, prefix="" + name + ".", keep_vars=False)
        for hook in self._state_dict_hooks.values():
            hook_result = hook(self, destination, "", local_metadata)
            if hook_result is not None:
                destination = hook_result
        return destination

    def load_state_dict(self, state_dict: dict[str, torch.Tensor]):
        missing_keys: list[str] = []
        unexpected_keys: list[str] = []
        error_msgs: list[str] = []

        metadata = getattr(state_dict, "_metadata", None)
        state_dict = state_dict.copy()
        if metadata is not None:
            state_dict["_metadata"] = metadata

        def load(module: torch.nn.Module, prefix=""):
            local_metadata = {} if metadata is None else metadata.get(prefix[:-1], {})
            module._load_from_state_dict(
                state_dict,
                prefix,
                local_metadata,
                True,
                missing_keys,
                unexpected_keys,
                error_msgs,
            )
            for name, child in module._modules.items():
                if child is not None:
                    if not isinstance(child, Network):
                        if isinstance(child, torch.nn.modules.conv._ConvNd) or isinstance(module, torch.nn.Linear):

                            current_size = child.weight.shape[0]
                            last_size = state_dict[prefix + name + ".weight"].shape[0]

                            if current_size != last_size:
                                print(
                                    f"Warning: The size of '{prefix + name}' has changed from {last_size}"
                                    f" to {current_size}. Please check for potential impacts"
                                )
                                ModuleArgsDict.init_func(child, self.init_type, self.init_gain)

                                with torch.no_grad():
                                    child.weight[:last_size] = state_dict[prefix + name + ".weight"]
                                    if child.bias is not None:
                                        child.bias[:last_size] = state_dict[prefix + name + ".bias"]
                                return
                        load(child, prefix + name + ".")

        load(self)

        if len(unexpected_keys) > 0:
            formatted_keys = ", ".join(f'"{k}"' for k in unexpected_keys)
            error_msgs.insert(
                0,
                f"Unexpected key(s) in state_dict: {formatted_keys}.",
            )
        if len(missing_keys) > 0:
            formatted_keys = ", ".join(f'"{k}"' for k in missing_keys)
            error_msgs.insert(
                0,
                f"Missing key(s) in state_dict: {formatted_keys}.",
            )

        if len(error_msgs) > 0:
            formatted_errors = "\n\t".join(error_msgs)
            raise RuntimeError(
                f"Error(s) in loading state_dict for {self.__class__.__name__}:\n\t{formatted_errors}",
            )

    def apply(self, fn: Callable[[torch.nn.Module], None]) -> None:
        for module in self.children():
            if not isinstance(module, Network):
                module.apply(fn)
        fn(self)

    @_function_network()
    def load(
        self,
        state_dict: dict[str, dict[str, torch.Tensor] | int],
        init: bool = True,
        ema: bool = False,
    ):
        if init:
            self.apply(
                partial(
                    ModuleArgsDict.init_func,
                    init_type=self.init_type,
                    init_gain=self.init_gain,
                )
            )
        name = "Model"
        if ema:
            if name + "_EMA" in state_dict:
                name += "_EMA"
        if name in state_dict:
            value = state_dict[name]
            model_state_dict_tmp = {}
            if isinstance(value, dict):
                model_state_dict_tmp = {k.split(".")[-1]: v for k, v in value.items()}[self.get_name()]
            modules_name = self.get_mapping()
            model_state_dict: OrderedDict[str, torch.Tensor] = OrderedDict()

            for alias in model_state_dict_tmp.keys():
                prefix = ".".join(alias.split(".")[:-1])
                alias_list = [
                    (".".join(prefix.split(".")[: len(i.split("."))]), v)
                    for i, v in modules_name.items()
                    if prefix.startswith(i)
                ]

                if len(alias_list):
                    for a, b in alias_list:
                        model_state_dict[alias.replace(a, b)] = model_state_dict_tmp[alias]
                        break
                else:
                    model_state_dict[alias] = model_state_dict_tmp[alias]
            self.load_state_dict(model_state_dict)
        if f"{self.get_name()}_optimizer_state_dict" in state_dict and self.optimizer:
            last_lr = self.optimizer.param_groups[0]["lr"]
            self.optimizer.load_state_dict(state_dict[f"{self.get_name()}_optimizer_state_dict"])
            self.optimizer.param_groups[0]["lr"] = last_lr
        if f"{self.get_name()}_it" in state_dict:
            _it = state_dict.get(f"{self.get_name()}_it")
            if isinstance(_it, int):
                self._it = _it
        if f"{self.get_name()}_nb_lr_update" in state_dict:
            _nb_lr_update = state_dict.get(f"{self.get_name()}_nb_lr_update")
            if isinstance(_nb_lr_update, int):
                self._nb_lr_update = _nb_lr_update

        for scheduler in self.schedulers:
            if scheduler.last_epoch == -1:
                scheduler.last_epoch = self._nb_lr_update
        self.initialized()

    def _compute_channels_trace(
        self,
        module: ModuleArgsDict,
        in_channels: int,
        gradient_checkpoints: list[str] | None,
        gpu_checkpoints: list[str] | None,
        name: str | None = None,
        in_is_channel: bool = True,
        out_channels: int | None = None,
        out_is_channel: bool = True,
    ) -> tuple[int, bool, int | None, bool]:

        for k1, v1 in module.items():
            if isinstance(v1, ModuleArgsDict):
                for t in module._modulesArgs[k1].out_branch:
                    last = None
                    for k2, _ in v1.items():
                        if t in v1._modulesArgs[k2].out_branch:
                            last = k2
                    if last is not None:
                        v1._modulesArgs[last]._isEnd = True
                    else:
                        v1._modulesArgs[k2]._isEnd = True

        for k, v in module.items():
            if hasattr(v, "in_channels"):
                if v.in_channels:
                    in_channels = v.in_channels
            if hasattr(v, "in_features"):
                if v.in_features:
                    in_channels = v.in_features
            key = name + "." + k if name else k

            if gradient_checkpoints:
                if key in gradient_checkpoints:
                    module._modulesArgs[k].isCheckpoint = True

            if gpu_checkpoints:
                if key in gpu_checkpoints:
                    module._modulesArgs[k].isGPU_Checkpoint = True

            module._modulesArgs[k].in_channels = in_channels
            module._modulesArgs[k].in_is_channel = in_is_channel

            if isinstance(v, ModuleArgsDict):
                in_channels, in_is_channel, out_channels, out_is_channel = self._compute_channels_trace(
                    v,
                    in_channels,
                    gradient_checkpoints,
                    gpu_checkpoints,
                    key,
                    in_is_channel,
                    out_channels,
                    out_is_channel,
                )

            if v.__class__.__name__ == "ToChannels":
                out_is_channel = True

            if v.__class__.__name__ == "ToFeatures":
                out_is_channel = False

            if hasattr(v, "out_channels"):
                if v.out_channels:
                    out_channels = v.out_channels
            if hasattr(v, "out_features"):
                if v.out_features:
                    out_channels = v.out_features

            module._modulesArgs[k].out_channels = out_channels
            module._modulesArgs[k].out_is_channel = out_is_channel

            in_channels = out_channels if out_channels is not None else in_channels
            in_is_channel = out_is_channel

        return in_channels, in_is_channel, out_channels, out_is_channel

    @_function_network()
    def init(self, autocast: bool, state: State, group_dest: list[str], key: str) -> None:
        if self.outputs_criterions_loader:
            self.measure = Measure(key, self.outputs_criterions_loader)
            self.measure.init(self, group_dest)
        if self.patch is not None:
            self.patch.init(f"{konfai_root()}.Model.{key}.Patch")
        if state != State.PREDICTION:
            self.scaler = torch.amp.GradScaler("cuda", enabled=autocast)
            if self.optimizerLoader:
                self.optimizer = self.optimizerLoader.get_optimizer(
                    key, self.parameters(state == State.TRANSFER_LEARNING)
                )
                self.optimizer.zero_grad()

            if self.lr_schedulers_loader and self.optimizer:
                for schedulers_classname, schedulers in self.lr_schedulers_loader.items():
                    self.schedulers[schedulers.getschedulers(key, schedulers_classname, self.optimizer)] = (
                        schedulers.nb_step
                    )

    def initialized(self):
        pass

    def named_forward(self, *inputs: torch.Tensor) -> Iterator[tuple[str, torch.Tensor]]:
        if self.patch:
            self.patch.load(inputs[0].shape[2:])
            accumulators: dict[str, Accumulator] = {}

            patch_iterator = self.patch.disassemble(*inputs)
            buffer = []
            for i, patch_input in enumerate(patch_iterator):
                for name, output_layer in super().named_forward(*patch_input):
                    yield f";accu;{name}", output_layer
                    buffer.append((name.split(".")[0], output_layer))
                    if len(buffer) == 2:
                        if buffer[0][0] != buffer[1][0]:
                            if self._modulesArgs[buffer[0][0]]._isEnd:
                                if buffer[0][0] not in accumulators:
                                    accumulators[buffer[0][0]] = Accumulator(
                                        self.patch.get_patch_slices(),
                                        self.patch.patch_size,
                                        self.patch.patch_combine,
                                    )
                                accumulators[buffer[0][0]].add_layer(i, buffer[0][1])
                        buffer.pop(0)
                if self._modulesArgs[buffer[0][0]]._isEnd:
                    if buffer[0][0] not in accumulators:
                        accumulators[buffer[0][0]] = Accumulator(
                            self.patch.get_patch_slices(),
                            self.patch.patch_size,
                            self.patch.patch_combine,
                        )
                    accumulators[buffer[0][0]].add_layer(i, buffer[0][1])
            for name, accumulator in accumulators.items():
                yield name, accumulator.assemble()
        else:
            for name, output_layer in super().named_forward(*inputs):
                yield name, output_layer

    def get_layers(
        self, inputs: list[torch.Tensor], layers_name: list[str]
    ) -> Iterator[tuple[str, torch.Tensor, PatchIndexed | None]]:
        layers_name = layers_name.copy()
        output_layer_accumulator: dict[str, Accumulator] = {}
        output_layer_patch_indexed: dict[str, PatchIndexed] = {}
        it = 0
        debug = "KONFAI_DEBUG" in os.environ
        for name_tmp, output_layer in self.named_forward(*inputs):
            name = name_tmp.replace(";accu;", "")
            if debug:
                if "KONFAI_DEBUG_LAST_LAYER" in os.environ:
                    os.environ["KONFAI_DEBUG_LAST_LAYER"] = f"{os.environ['KONFAI_DEBUG_LAST_LAYER']}|{name}:"
                    f"{get_gpu_memory(output_layer.device)}:"
                    f"{str(output_layer.device).replace('cuda:', '')}"
                else:
                    os.environ["KONFAI_DEBUG_LAST_LAYER"] = (
                        f"{name}:{get_gpu_memory(output_layer.device)}:{str(output_layer.device).replace('cuda:', '')}"
                    )
            it += 1
            if name in layers_name or name_tmp in layers_name:
                if ";accu;" in name_tmp:
                    if name not in output_layer_patch_indexed:
                        network_name = (
                            name_tmp.split(".;accu;")[-2].split(".")[-1]
                            if ".;accu;" in name_tmp
                            else name_tmp.split(";accu;")[-2].split(".")[-1]
                        )
                        module = self
                        network = None
                        if network_name == "":
                            network = module
                        else:
                            for n in name.split("."):
                                module = module[n]
                                if isinstance(module, Network) and n == network_name:
                                    network = module
                                    break

                        if network and network.patch:
                            output_layer_patch_indexed[name] = PatchIndexed(network.patch, 0)

                    if name not in output_layer_accumulator:
                        output_layer_accumulator[name] = Accumulator(
                            output_layer_patch_indexed[name].patch.get_patch_slices(0),
                            output_layer_patch_indexed[name].patch.patch_size,
                            output_layer_patch_indexed[name].patch.patch_combine,
                        )

                    if name_tmp in layers_name:
                        output_layer_accumulator[name].add_layer(output_layer_patch_indexed[name].index, output_layer)
                        output_layer_patch_indexed[name].index += 1
                        if output_layer_accumulator[name].is_full():
                            output_layer = output_layer_accumulator[name].assemble()
                            output_layer_accumulator.pop(name)
                            output_layer_patch_indexed.pop(name)
                            layers_name.remove(name_tmp)
                            yield name_tmp, output_layer, None

                if name in layers_name:
                    if ";accu;" in name_tmp:
                        yield name, output_layer, output_layer_patch_indexed[name]
                        output_layer_patch_indexed[name].index += 1
                        if output_layer_patch_indexed[name].is_full():
                            output_layer_patch_indexed.pop(name)
                            layers_name.remove(name)
                    else:
                        layers_name.remove(name)
                        yield name, output_layer, None

            if not len(layers_name):
                break

    def init_outputs_group(self):
        metric_tmp = {
            network.measure: network.measure.outputs_criterions.keys()
            for network in self.get_networks().values()
            if network.measure
        }
        for k, v in metric_tmp.items():
            for a in v:
                outputs_group = OutputsGroup(k)
                outputs_group.append(a)
                for targets_group in k.outputs_criterions[a].keys():
                    if ":" in targets_group:
                        outputs_group.append(targets_group.replace(":", "."))

                self.outputsGroup.append(outputs_group)

    def forward(
        self,
        data_dict: dict[tuple[str, bool], torch.Tensor],
        output_layers: list[str] = [],
    ) -> list[tuple[str, torch.Tensor]]:
        if not len(self.outputsGroup) and not len(output_layers):
            return []

        self.reset_loss()
        results = []
        measure_output_layers = set()
        for _outputs_group in self.outputsGroup:
            for name in _outputs_group:
                measure_output_layers.add(name)
        for name, layer, patch_indexed in self.get_layers(
            [v for k, v in data_dict.items() if k[1]],
            list(set(list(measure_output_layers) + output_layers)),
        ):

            outputs_group = [outputs_group for outputs_group in self.outputsGroup if name in outputs_group]
            if len(outputs_group) > 0:
                if patch_indexed is None:
                    targets = {k[0]: v for k, v in data_dict.items()}
                    nb = 1
                else:
                    targets = {
                        k[0]: patch_indexed.patch.get_data(v, patch_indexed.index, 0, False)
                        for k, v in data_dict.items()
                    }
                    nb = patch_indexed.patch.get_size(0)

                for output_group in outputs_group:
                    output_group.add_layer(name, layer)
                    if output_group.is_done():
                        targets.update(
                            {k.replace(".", ":"): v for k, v in output_group.layers.items() if k != output_group[0]}
                        )
                        output_group.measure.update(
                            output_group[0],
                            output_group.layers[output_group[0]],
                            targets,
                            self._it,
                            nb,
                            self.training,
                        )
                        output_group.clear()
            if name in output_layers:
                results.append((name, layer))
        return results

    @_function_network()
    def reset_loss(self):
        if self.measure:
            self.measure.reset_loss()

    @_function_network()
    def backward(self, model: Self):
        if self.measure:
            if self.scaler and self.optimizer:
                model._requires_grad(list(self.measure.outputs_criterions.keys()))
                for loss in self.measure.get_loss():
                    self.scaler.scale(loss / self.nb_batch_per_step).backward()

                    if self._it % self.nb_batch_per_step == 0:
                        self.scaler.step(self.optimizer)
                        self.scaler.update()
                        self.optimizer.zero_grad(set_to_none=True)
                self._it += 1

    @_function_network()
    def update_lr(self):
        self._nb_lr_update += 1
        step = 0
        _scheduler = None
        for _scheduler, value in self.schedulers.items():
            if value is None or (self._nb_lr_update >= step and self._nb_lr_update < step + value):
                break
            step += value
        if _scheduler:
            if _scheduler.__class__.__name__ == "ReduceLROnPlateau":
                if self.measure:
                    _scheduler.step(sum(self.measure.get_last_values(0).values()))
            else:
                _scheduler.step()

    @_function_network()
    def get_networks(self) -> Self:
        return self

    @staticmethod
    def to(module: ModuleArgsDict, device: int):
        if "device" not in os.environ:
            os.environ["device"] = str(device)
        for k, v in module.items():
            if module._modulesArgs[k].gpu == "cpu":
                if module._modulesArgs[k].isGPU_Checkpoint:
                    os.environ["device"] = str(int(os.environ["device"]) + 1)
                module._modulesArgs[k].gpu = str(get_device(int(os.environ["device"])))
                if isinstance(v, ModuleArgsDict):
                    v = Network.to(v, int(os.environ["device"]))
                else:
                    v = v.to(get_device(int(os.environ["device"])))
        if isinstance(module, Network):
            if module.optimizer is not None:
                for state in module.optimizer.state.values():
                    for k, v in state.items():
                        if isinstance(v, torch.Tensor):
                            state[k] = v.to(get_device(int(os.environ["device"])))
        return module

    def get_name(self) -> str:
        return self.name

    def set_name(self, name: str) -> Self:
        self.name = name
        return self

    def set_state(self, state: NetState):
        for module in self.modules():
            if isinstance(module, ModuleArgsDict):
                module._training = state


class MinimalModel(Network):

    def __init__(
        self,
        model: Network,
        optimizer: OptimizerLoader = OptimizerLoader(),
        schedulers: dict[str, LRSchedulersLoader] = {"default:ReduceLROnPlateau": LRSchedulersLoader(0)},
        outputs_criterions: dict[str, TargetCriterionsLoader] = {"default": TargetCriterionsLoader()},
        patch: ModelPatch | None = None,
        dim: int = 3,
        nb_batch_per_step=1,
        init_type="normal",
        init_gain=0.02,
    ):
        super().__init__(
            1,
            optimizer,
            schedulers,
            outputs_criterions,
            patch,
            nb_batch_per_step,
            init_type,
            init_gain,
            dim,
        )
        self.add_module("Model", model)


class ModelLoader:

    @config("Model")
    def __init__(self, classpath: str = "default:segmentation.UNet.UNet") -> None:
        self.module, self.name = get_module(classpath, "konfai.models")

    def get_model(
        self,
        train: bool = True,
        konfai_args: str | None = None,
        konfai_without=[
            "optimizer",
            "schedulers",
            "nb_batch_per_step",
            "init_type",
            "init_gain",
        ],
    ) -> Network:
        if not konfai_args:
            konfai_args = f"{konfai_root()}.Model"
        konfai_args += "." + self.name
        model = config(konfai_args)(getattr(importlib.import_module(self.module), self.name))(
            config=None, konfai_without=konfai_without if not train else []
        )
        if not isinstance(model, Network):
            model = config(konfai_args)(partial(MinimalModel, model))(
                config=None, konfai_without=konfai_without + ["model"] if not train else []
            )
            model.set_name(self.name)
        return model


class CPUModel:

    def __init__(self, model: Network) -> None:
        self.module = model
        self.device = torch.device("cpu")

    def train(self):
        self.module.train()

    def eval(self):  # noqa: A003
        self.module.eval()

    def __call__(
        self,
        data_dict: dict[tuple[str, bool], torch.Tensor],
        output_layers: list[str] = [],
    ) -> list[tuple[str, torch.Tensor]]:
        return self.module(data_dict, output_layers)



import importlib
import os
from abc import ABC, abstractmethod

import numpy as np
import SimpleITK as sitk  # noqa: N813
import torch
import torch.nn.functional as F  # noqa: N812

from konfai import konfai_root
from konfai.utils.config import config
from konfai.utils.dataset import Attribute, Dataset, data_to_image
from konfai.utils.utils import AugmentationError, NeedDevice, get_module


def _translate_2d_matrix(t: torch.Tensor) -> torch.Tensor:
    return torch.cat(
        (
            torch.cat((torch.eye(2), torch.tensor([[t[0]], [t[1]]])), dim=1),
            torch.Tensor([[0, 0, 1]]),
        ),
        dim=0,
    )


def _translate_3d_matrix(t: torch.Tensor) -> torch.Tensor:
    return torch.cat(
        (
            torch.cat((torch.eye(3), torch.tensor([[t[0]], [t[1]], [t[2]]])), dim=1),
            torch.Tensor([[0, 0, 0, 1]]),
        ),
        dim=0,
    )


def _scale_2d_matrix(s: torch.Tensor) -> torch.Tensor:
    return torch.cat(
        (
            torch.cat((torch.eye(2) * s, torch.tensor([[0], [0]])), dim=1),
            torch.tensor([[0, 0, 1]]),
        ),
        dim=0,
    )


def _scale_3d_matrix(s: torch.Tensor) -> torch.Tensor:
    return torch.cat(
        (
            torch.cat((torch.eye(3) * s, torch.tensor([[0], [0], [0]])), dim=1),
            torch.tensor([[0, 0, 0, 1]]),
        ),
        dim=0,
    )


def _rotation_3d_matrix(rotation: torch.Tensor, center: torch.Tensor | None = None) -> torch.Tensor:
    a = torch.tensor(
        [
            [torch.cos(rotation[2]), -torch.sin(rotation[2]), 0],
            [torch.sin(rotation[2]), torch.cos(rotation[2]), 0],
            [0, 0, 1],
        ]
    )
    b = torch.tensor(
        [
            [torch.cos(rotation[1]), 0, torch.sin(rotation[1])],
            [0, 1, 0],
            [-torch.sin(rotation[1]), 0, torch.cos(rotation[1])],
        ]
    )
    c = torch.tensor(
        [
            [1, 0, 0],
            [0, torch.cos(rotation[0]), -torch.sin(rotation[0])],
            [0, torch.sin(rotation[0]), torch.cos(rotation[0])],
        ]
    )
    rotation_matrix = torch.cat(
        (
            torch.cat((a.mm(b).mm(c), torch.zeros((3, 1))), dim=1),
            torch.tensor([[0, 0, 0, 1]]),
        ),
        dim=0,
    )
    if center is not None:
        translation_before = torch.eye(4)
        translation_before[:-1, -1] = -center
        rotation_matrix = translation_before.mm(rotation_matrix)
    if center is not None:
        translation_after = torch.eye(4)
        translation_after[:-1, -1] = center
        rotation_matrix = rotation_matrix.mm(translation_after)
    return rotation_matrix


def _rotation_2d_matrix(rotation: torch.Tensor, center: torch.Tensor | None = None) -> torch.Tensor:
    return torch.cat(
        (
            torch.cat(
                (
                    torch.tensor(
                        [
                            [torch.cos(rotation[0]), -torch.sin(rotation[0])],
                            [torch.sin(rotation[0]), torch.cos(rotation[0])],
                        ]
                    ),
                    torch.zeros((2, 1)),
                ),
                dim=1,
            ),
            torch.tensor([[0, 0, 1]]),
        ),
        dim=0,
    )


class Prob:

    @config()
    def __init__(self, prob: float = 1.0) -> None:
        self.prob = prob


class DataAugmentationsList:

    @config()
    def __init__(
        self,
        nb: int = 10,
        data_augmentations: dict[str, Prob] = {"default:Flip": Prob(1)},
    ) -> None:
        self.nb = nb
        self.data_augmentations: list[DataAugmentation] = []
        self.data_augmentationsLoader = data_augmentations

    def load(self, key: str, datasets: list[Dataset]):
        for augmentation, prob in self.data_augmentationsLoader.items():
            module, name = get_module(augmentation, "konfai.data.augmentation")
            data_augmentation: DataAugmentation = config(
                f"{konfai_root()}.Dataset.augmentations.{key}.data_augmentations.{augmentation}"
            )(getattr(importlib.import_module(module), name))(config=None)
            data_augmentation.load(prob.prob)
            data_augmentation.set_datasets(datasets)
            self.data_augmentations.append(data_augmentation)


class DataAugmentation(NeedDevice, ABC):

    def __init__(self, groups: list[str] | None = None) -> None:
        self.who_index: dict[int, list[int]] = {}
        self.shape_index: dict[int, list[list[int]]] = {}
        self._prob: float = 0
        self.groups = groups
        self.datasets: list[Dataset] = []

    def load(self, prob: float):
        self._prob = prob

    def set_datasets(self, datasets: list[Dataset]):
        self.datasets = datasets

    def state_init(
        self,
        index: None | int,
        shapes: list[list[int]],
        caches_attribute: list[Attribute],
    ) -> list[list[int]]:
        if index is not None:
            if index not in self.who_index:
                self.who_index[index] = torch.where(torch.rand(len(shapes)) < self._prob)[0].tolist()
            else:
                return self.shape_index[index]
        else:
            index = 0
            self.who_index[index] = torch.where(torch.rand(len(shapes)) < self._prob)[0].tolist()

        if len(self.who_index[index]) > 0:
            for i, shape in enumerate(
                self._state_init(
                    index,
                    [shapes[i] for i in self.who_index[index]],
                    [caches_attribute[i] for i in self.who_index[index]],
                )
            ):
                shapes[self.who_index[index][i]] = shape
        self.shape_index[index] = shapes
        return self.shape_index[index]

    @abstractmethod
    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        pass

    def __call__(
        self,
        name: str,
        index: int,
        tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        if len(self.who_index[index]) > 0:
            for i, result in enumerate(self._compute(name, index, [tensors[i] for i in self.who_index[index]])):
                tensors[self.who_index[index][i]] = result
        return tensors

    @abstractmethod
    def _compute(
        self,
        name: str,
        index: int,
        tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        pass

    def inverse(self, index: int, a: int, tensor: torch.Tensor) -> torch.Tensor:
        if a in self.who_index[index]:
            tensor = self._inverse(index, a, tensor)
        return tensor

    @abstractmethod
    def _inverse(self, index: int, a: int, tensor: torch.Tensor) -> torch.Tensor:
        pass


class EulerTransform(DataAugmentation):

    def __init__(self) -> None:
        super().__init__()
        self.matrix: dict[int, list[torch.Tensor]] = {}

    def _compute(
        self,
        name: str,
        index: int,
        tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        results = []
        for tensor, matrix in zip(tensors, self.matrix[index]):
            results.append(
                F.grid_sample(
                    tensor.unsqueeze(0).type(torch.float32),
                    F.affine_grid(matrix[:, :-1, ...], [1] + list(tensor.shape), align_corners=True).to(tensor.device),
                    align_corners=True,
                    mode="bilinear",
                    padding_mode="reflection",
                )
                .type(tensor.dtype)
                .squeeze(0)
            )
        return results

    def _inverse(self, index: int, a: int, tensor: torch.Tensor) -> torch.Tensor:
        return (
            F.grid_sample(
                tensor.unsqueeze(0).type(torch.float32),
                F.affine_grid(
                    self.matrix[index][a].inverse()[:, :-1, ...],
                    [1] + list(tensor.shape),
                    align_corners=True,
                ).to(tensor.device),
                align_corners=True,
                mode="bilinear",
                padding_mode="reflection",
            )
            .type(tensor.dtype)
            .squeeze(0)
        )


class Translate(EulerTransform):

    def __init__(self, t_min: float = -10, t_max=10, is_int: bool = False):
        super().__init__()
        self.t_min = t_min
        self.t_max = t_max
        self.is_int = is_int

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        dim = len(shapes[0])
        func = _translate_3d_matrix if dim == 3 else _translate_2d_matrix
        translate = torch.rand((len(shapes), dim)) * torch.tensor(self.t_max - self.t_min) + torch.tensor(self.t_min)
        if self.is_int:
            translate = torch.round(translate * 100) / 100
        self.matrix[index] = [torch.unsqueeze(func(value), dim=0) for value in translate]
        return shapes


class Rotate(EulerTransform):

    def __init__(self, a_min: float = 0, a_max: float = 360, is_quarter: bool = False):
        super().__init__()
        self.a_min = a_min
        self.a_max = a_max
        self.is_quarter = is_quarter

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        dim = len(shapes[0])
        func = _rotation_3d_matrix if dim == 3 else _rotation_2d_matrix
        angles = []

        if self.is_quarter:
            angles = torch.Tensor.repeat(torch.tensor([90, 180, 270]), 3)
        else:
            angles = torch.rand((len(shapes), dim)) * torch.tensor(self.a_max - self.a_min) + torch.tensor(self.a_min)

        self.matrix[index] = [torch.unsqueeze(func(value), dim=0) for value in angles]
        return shapes


class Scale(EulerTransform):

    def __init__(self, s_std: float = 0.2):
        super().__init__()
        self.s_std = s_std

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        func = _scale_3d_matrix if len(shapes[0]) == 3 else _scale_2d_matrix
        scale = torch.Tensor.repeat(
            torch.exp2(torch.randn(len(shapes)) * self.s_std).unsqueeze(1),
            [1, len(shapes[0])],
        )
        self.matrix[index] = [torch.unsqueeze(func(value), dim=0) for value in scale]
        return shapes


class Flip(DataAugmentation):

    def __init__(self, f_prob: list[float] = [0.33, 0.33, 0.33]) -> None:
        super().__init__()
        self.f_prob = f_prob
        self.flip: dict[int, list[int]] = {}

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        prob = torch.rand((len(shapes), len(self.f_prob))) < torch.tensor(self.f_prob)
        dims = torch.tensor([1, 2, 3][: len(self.f_prob)])
        self.flip[index] = [dims[mask].tolist() for mask in prob]
        return shapes

    def _compute(
        self,
        name: str,
        index: int,
        tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        results = []
        for tensor, flip in zip(tensors, self.flip[index]):
            results.append(torch.flip(tensor, dims=flip))
        return results

    def _inverse(self, index: int, a: int, tensor: torch.Tensor) -> torch.Tensor:
        return torch.flip(tensor, dims=self.flip[index][a])


class ColorTransform(DataAugmentation):

    def __init__(self, groups: list[str] | None = None) -> None:
        super().__init__(groups)
        self.matrix: dict[int, list[torch.Tensor]] = {}

    def _compute(
        self,
        name: str,
        index: int,
        tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        results = []
        for tensor, matrix in zip(tensors, self.matrix[index]):
            result = tensor.reshape([*tensor.shape[:1], int(np.prod(tensor.shape[1:]))])
            if tensor.shape[0] == 3:
                matrix = matrix.to(tensor.device)
                result = matrix[:, :3, :3] @ result.float() + matrix[:, :3, 3:]
            elif tensor.shape[0] == 1:
                matrix = matrix[:, :3, :].mean(dim=1, keepdims=True).to(tensor.device)
                result = result.float() * matrix[:, :, :3].sum(dim=2, keepdims=True) + matrix[:, :, 3:]
            else:
                raise AugmentationError("Image must be RGB (3 channels) or L (1 channel)")
            results.append(result.reshape(tensor.shape))
        return results

    def _inverse(self, index: int, a: int, tensors: torch.Tensor) -> torch.Tensor:
        pass


class Brightness(ColorTransform):

    def __init__(self, b_std: float, groups: list[str] | None = None) -> None:
        super().__init__(groups)
        self.b_std = b_std

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        brightness = torch.Tensor.repeat((torch.randn(len(shapes)) * self.b_std).unsqueeze(1), [1, 3])
        self.matrix[index] = [torch.unsqueeze(_translate_3d_matrix(value), dim=0) for value in brightness]
        return shapes


class Contrast(ColorTransform):

    def __init__(self, c_std: float, groups: list[str] | None = None) -> None:
        super().__init__(groups)
        self.c_std = c_std

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        contrast = torch.exp2(torch.randn(len(shapes)) * self.c_std)
        self.matrix[index] = [torch.unsqueeze(_scale_3d_matrix(value), dim=0) for value in contrast]
        return shapes


class LumaFlip(ColorTransform):

    def __init__(self, groups: list[str] | None = None) -> None:
        super().__init__(groups)
        self.v = torch.tensor([1, 1, 1, 0]) / torch.sqrt(torch.tensor(3))

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        luma = torch.floor(torch.rand([len(shapes), 1, 1]) * 2)
        self.matrix[index] = [torch.unsqueeze((torch.eye(4) - 2 * self.v.ger(self.v) * value), dim=0) for value in luma]
        return shapes


class HUE(ColorTransform):

    def __init__(self, hue_max: float, groups: list[str] | None = None) -> None:
        super().__init__(groups)
        self.hue_max = hue_max
        self.v = torch.tensor([1, 1, 1]) / torch.sqrt(torch.tensor(3))

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        theta = (torch.rand([len(shapes)]) * 2 - 1) * np.pi * self.hue_max
        self.matrix[index] = [torch.unsqueeze(_rotation_3d_matrix(value.repeat(3), self.v), dim=0) for value in theta]
        return shapes


class Saturation(ColorTransform):

    def __init__(self, s_std: float, groups: list[str] | None = None) -> None:
        super().__init__(groups)
        self.s_std = s_std
        self.v = torch.tensor([1, 1, 1, 0]) / torch.sqrt(torch.tensor(3))

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        saturation = torch.exp2(torch.randn(len(shapes)) * self.s_std)
        self.matrix[index] = [
            (self.v.ger(self.v) + (torch.eye(4) - self.v.ger(self.v))).unsqueeze(0) * value for value in saturation
        ]
        return shapes


class Noise(DataAugmentation):

    def __init__(
        self,
        n_std: float,
        noise_step: int = 1000,
        beta_start: float = 1e-4,
        beta_end: float = 0.02,
        groups: list[str] | None = None,
    ) -> None:
        super().__init__(groups)
        self.n_std = n_std
        self.noise_step = noise_step

        self.ts: dict[int, list[torch.Tensor]] = {}
        self.betas = torch.linspace(beta_start, beta_end, noise_step)
        self.betas = Noise.enforce_zero_terminal_snr(self.betas)
        self.alphas = 1 - self.betas
        self.alpha_hat = torch.concat((torch.ones(1), torch.cumprod(self.alphas, dim=0)))
        self.max_T = 0.0

        self.C = 1
        self.n = 4
        self.d = 0.25
        self._prob = 1

    @staticmethod
    def enforce_zero_terminal_snr(betas: torch.Tensor):
        alphas = 1 - betas
        alphas_bar = alphas.cumprod(0)
        alphas_bar_sqrt = alphas_bar.sqrt()
        alphas_bar_sqrt_0 = alphas_bar_sqrt[0].clone()
        alphas_bar_sqrt_t = alphas_bar_sqrt[-1].clone()
        alphas_bar_sqrt -= alphas_bar_sqrt_t
        alphas_bar_sqrt *= alphas_bar_sqrt_0 / (alphas_bar_sqrt_0 - alphas_bar_sqrt_t)
        alphas_bar = alphas_bar_sqrt**2
        alphas = alphas_bar[1:] / alphas_bar[:-1]
        alphas = torch.cat([alphas_bar[0:1], alphas])
        betas = 1 - alphas
        return betas

    def load(self, prob: float):
        self.max_T = prob * self.noise_step

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        if int(self.max_T) == 0:
            self.ts[index] = [0 for _ in shapes]
        else:
            self.ts[index] = [torch.randint(0, int(self.max_T), (1,)) for _ in shapes]
        return shapes

    def _compute(
        self,
        name: str,
        index: int,
        tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        results = []
        for tensor, t in zip(tensors, self.ts[index]):
            alpha_hat_t = self.alpha_hat[t].to(tensor.device).reshape(*[1 for _ in range(len(tensor.shape))])
            results.append(
                alpha_hat_t.sqrt() * tensor
                + (1 - alpha_hat_t).sqrt() * torch.randn_like(tensor.float()).to(tensor.device) * self.n_std
            )
        return results

    def _inverse(self, index: int, a: int, tensor: torch.Tensor) -> torch.Tensor:
        pass


class CutOUT(DataAugmentation):

    def __init__(
        self,
        c_prob: float,
        cutout_size: int,
        value: float,
        groups: list[str] | None = None,
    ) -> None:
        super().__init__(groups)
        self.c_prob = c_prob
        self.cutout_size = cutout_size
        self.centers: dict[int, list[torch.Tensor]] = {}
        self.value = value

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        self.centers[index] = [torch.rand((3) if len(shape) == 3 else (2)) for shape in shapes]
        return shapes

    def _compute(
        self,
        name: str,
        index: int,
        tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        results = []
        for tensor, center in zip(tensors, self.centers[index]):
            masks = []
            for i, w in enumerate(tensor.shape[1:]):
                re = [1] * i + [-1] + [1] * (len(tensor.shape[1:]) - i - 1)
                masks.append(
                    ((torch.arange(w).reshape(re) + 0.5) / w - center[i].reshape([1, 1])).abs()
                    >= torch.tensor(self.cutout_size).reshape([1, 1]) / 2
                )
            result = masks[0]
            for mask in masks[1:]:
                result = torch.logical_or(result, mask)
            result = result.unsqueeze(0).repeat([tensor.shape[0], *[1 for _ in range(len(tensor.shape) - 1)]])
            results.append(
                torch.where(
                    result.to(tensor.device) == 1,
                    tensor,
                    torch.tensor(self.value).to(tensor.device),
                )
            )
        return results

    def _inverse(self, index: int, a: int, tensor: torch.Tensor) -> torch.Tensor:
        pass


class Elastix(DataAugmentation):

    def __init__(self, grid_spacing: int = 16, max_displacement: int = 16) -> None:
        super().__init__()
        self.grid_spacing = grid_spacing
        self.max_displacement = max_displacement
        self.displacement_fields: dict[int, list[torch.Tensor]] = {}
        self.displacement_fields_true: dict[int, list[torch.Tensor]] = {}

    @staticmethod
    def _format_loc(new_locs, shape):
        for i in range(len(shape)):
            new_locs[..., i] = 2 * (new_locs[..., i] / (shape[i] - 1) - 0.5)
        new_locs = new_locs[..., list(reversed(range(len(shape))))]
        return new_locs

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        print(f"Compute Displacement Field for index {index}")
        self.displacement_fields[index] = []
        self.displacement_fields_true[index] = []
        for i, (shape, cache_attribute) in enumerate(zip(shapes, caches_attribute)):
            shape = shape
            dim = len(shape)
            if "Spacing" not in cache_attribute:
                spacing = np.array([1.0 for _ in range(dim)])
            else:
                spacing = cache_attribute.get_np_array("Spacing")

            grid_physical_spacing = [self.grid_spacing] * dim
            image_physical_size = [size * spacing for size, spacing in zip(shape, spacing)]
            mesh_size = [
                int(image_size / grid_spacing + 0.5)
                for image_size, grid_spacing in zip(image_physical_size, grid_physical_spacing)
            ]
            if "Spacing" not in cache_attribute:
                cache_attribute["Spacing"] = np.array([1.0 for _ in range(dim)])
            if "Origin" not in cache_attribute:
                cache_attribute["Origin"] = np.array([1.0 for _ in range(dim)])
            if "Direction" not in cache_attribute:
                cache_attribute["Direction"] = np.eye(dim).flatten()

            ref_image = data_to_image(np.expand_dims(np.zeros(shape), 0), cache_attribute)

            bspline_transform = sitk.BSplineTransformInitializer(
                image1=ref_image, transformDomainMeshSize=mesh_size, order=3
            )
            displacement_filter = sitk.TransformToDisplacementFieldFilter()
            displacement_filter.SetReferenceImage(ref_image)

            vectors = [torch.arange(0, s) for s in shape]
            grids = torch.meshgrid(vectors, indexing="ij")
            grid = torch.stack(grids)
            grid = torch.unsqueeze(grid, 0)
            grid = grid.type(torch.float).permute([0] + [i + 2 for i in range(len(shape))] + [1])

            control_points = torch.rand(*[size + 3 for size in mesh_size], dim)
            control_points -= 0.5
            control_points *= 2 * self.max_displacement
            bspline_transform.SetParameters(control_points.flatten().tolist())
            displacement = sitk.GetArrayFromImage(displacement_filter.Execute(bspline_transform))
            self.displacement_fields_true[index].append(displacement)
            new_locs = grid + torch.unsqueeze(torch.from_numpy(displacement), 0).type(torch.float32)
            self.displacement_fields[index].append(Elastix._format_loc(new_locs, shape))
            print(f"Compute in progress : {(i + 1) / len(shapes) * 100:.2f} %")
        return shapes

    def _compute(
        self,
        name: str,
        index: int,
        tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        results = []
        for tensor, displacement_field in zip(tensors, self.displacement_fields[index]):
            results.append(
                F.grid_sample(
                    tensor.type(torch.float32).unsqueeze(0),
                    displacement_field.to(tensor.device),
                    align_corners=True,
                    mode="bilinear",
                    padding_mode="border",
                )
                .type(tensor.dtype)
                .squeeze(0)
            )
        return results

    def _inverse(self, index: int, a: int, tensor: torch.Tensor) -> torch.Tensor:
        pass


class Permute(DataAugmentation):

    def __init__(self, prob_permute: list[float] | None = [0.5, 0.5]) -> None:
        super().__init__()
        self._permute_dims = torch.tensor([[0, 2, 1, 3], [0, 3, 1, 2]])
        self.prob_permute = prob_permute
        self.permute: dict[int, torch.Tensor] = {}

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        if len(shapes):
            dim = len(shapes[0])
            if dim != 3:
                raise ValueError("The permute augmentation only support 3D images")
            if self.prob_permute:
                if len(self.prob_permute) != 2:
                    raise ValueError("Size of prob_permute must be equal 2")
                self.permute[index] = torch.rand((len(shapes), len(self.prob_permute))) < torch.tensor(
                    self.prob_permute
                )
            else:
                if len(shapes) != 2:
                    raise ValueError("The number of augmentation images must be equal to 2")
                self.permute[index] = torch.eye(2, dtype=torch.bool)
            for i, prob in enumerate(self.permute[index]):
                for permute in self._permute_dims[prob]:
                    shapes[i] = [shapes[i][dim - 1] for dim in permute[1:]]
        return shapes

    def _compute(
        self,
        name: str,
        index: int,
        tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        results = []
        for tensor, prob in zip(tensors, self.permute[index]):
            res = tensor
            for permute in self._permute_dims[prob]:
                res = res.permute(tuple(permute))
            results.append(res)
        return results

    def _inverse(self, index: int, a: int, tensor: torch.Tensor) -> torch.Tensor:
        for permute in reversed(self._permute_dims[self.permute[index][a]]):
            tensor = tensor.permute(tuple(np.argsort(permute)))
        return tensor


class Mask(DataAugmentation):

    def __init__(self, mask: str, value: float, groups: list[str] | None = None) -> None:
        super().__init__(groups)
        if mask is not None:
            if os.path.exists(mask):
                self.mask = torch.tensor(sitk.GetArrayFromImage(sitk.ReadImage(mask)))
            else:
                raise NameError("Mask file not found")
        self.positions: dict[int, list[torch.Tensor]] = {}
        self.value = value

    def _state_init(self, index: int, shapes: list[list[int]], caches_attribute: list[Attribute]) -> list[list[int]]:
        self.positions[index] = [
            torch.rand((3) if len(shape) == 3 else (2))
            * (torch.tensor([max(s1 - s2, 0) for s1, s2 in zip(torch.tensor(shape), torch.tensor(self.mask.shape))]))
            for shape in shapes
        ]
        return [self.mask.shape for _ in shapes]

    def _compute(
        self,
        name: str,
        index: int,
        tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        results = []
        for tensor, position in zip(tensors, self.positions[index]):
            slices = [slice(None, None)] + [slice(int(s1), int(s1) + s2) for s1, s2 in zip(position, self.mask.shape)]
            padding = []
            for s1, s2 in zip(reversed(tensor.shape), reversed(self.mask.shape)):
                if s1 < s2:
                    pad = s2 - s1
                else:
                    pad = 0
                padding.append(0)
                padding.append(pad)
            value = (
                torch.tensor(0, dtype=torch.uint8)
                if tensor.dtype == torch.uint8
                else torch.tensor(self.value).to(tensor.device)
            )
            results.append(
                torch.where(
                    self.mask.to(tensor.device) == 1,
                    torch.nn.functional.pad(tensor, tuple(padding), mode="constant", value=value)[tuple(slices)],
                    value,
                )
            )
        return results

    def _inverse(self, index: int, a: int, tensor: torch.Tensor) -> torch.Tensor:
        pass



import torch
import torch.nn.functional as F  # noqa: N812

from konfai.data.patching import ModelPatch
from konfai.network import blocks, network

"""
"convnext_tiny_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_tiny_1k_224_ema.pth",
depths=[3, 3, 9, 3], dims=[96, 192, 384, 768]
"convnext_small_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_small_1k_224_ema.pth",
depths=[3, 3, 27, 3], dims=[96, 192, 384, 768]
"convnext_base_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_base_1k_224_ema.pth",
depths=[3, 3, 27, 3], dims=[128, 256, 512, 1024]
"convnext_large_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_large_1k_224_ema.pth",
[3, 3, 27, 3], dims=[192, 384, 768, 1536]
"convnext_tiny_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_tiny_22k_224.pth",
"convnext_small_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_small_22k_224.pth",
"convnext_base_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_base_22k_224.pth",
"convnext_large_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_large_22k_224.pth",
"convnext_xlarge_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_xlarge_22k_224.pth",
depths=[3, 3, 27, 3], dims=[256, 512, 1024, 2048]
"""


class LayerNorm(torch.nn.Module):

    def __init__(
        self,
        normalized_shape: int,
        eps: float = 1e-6,
        data_format: str = "channels_last",
    ):
        super().__init__()
        self.weight = torch.nn.parameter.Parameter(torch.ones(normalized_shape))
        self.bias = torch.nn.parameter.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (normalized_shape,)

    def forward(self, x):
        if self.data_format == "channels_last":
            return F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        elif self.data_format == "channels_first":
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            if len(x.shape) == 3:
                x = self.weight[:, None] * x + self.bias[:, None]
            elif len(x.shape) == 4:
                x = self.weight[:, None, None] * x + self.bias[:, None, None]
            else:
                x = self.weight[:, None, None, None] * x + self.bias[:, None, None, None]
            return x

    def extra_repr(self):
        return f"normalized_shape={self.normalized_shape}, eps={self.eps}, data_format={self.data_format})"


class DropPath(torch.nn.Module):

    def __init__(self, drop_prob: float = 0.0, scale_by_keep: bool = True):
        super().__init__()
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def forward(self, x):
        if self.drop_prob == 0.0 or not self.training:
            return x
        keep_prob = 1 - self.drop_prob
        shape = (x.shape[0],) + (1,) * (x.ndim - 1)
        random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
        if keep_prob > 0.0 and self.scale_by_keep:
            random_tensor.div_(keep_prob)
        return x * random_tensor

    def extra_repr(self):
        return f"drop_prob={round(self.drop_prob, 3)}"


class LayerScaler(torch.nn.Module):

    def __init__(self, init_value: float, dimensions: int):
        super().__init__()
        self.init_value = init_value
        self.gamma = torch.nn.Parameter(torch.ones(dimensions, 1, 1) * init_value)

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return self.gamma * tensor

    def extra_repr(self):
        return f"init_value={self.init_value}"


class BottleNeckBlock(network.ModuleArgsDict):
    def __init__(self, features: int, drop_p: float, layer_scaler_init_value: float, dim: int):
        super().__init__()
        self.add_module(
            "Conv_0",
            blocks.get_torch_module("Conv", dim)(features, features, kernel_size=7, padding=3, groups=features),
            alias=["dwconv"],
        )
        self.add_module("ToFeatures", blocks.ToFeatures(dim))
        self.add_module("LayerNorm", LayerNorm(features, eps=1e-6), alias=["norm"])
        self.add_module("Linear_1", torch.nn.Linear(features, features * 4), alias=["pwconv1"])
        self.add_module("GELU", torch.nn.GELU())
        self.add_module("Linear_2", torch.nn.Linear(features * 4, features), alias=["pwconv2"])
        self.add_module("ToChannels", blocks.ToChannels(dim))
        self.add_module(
            "LayerScaler",
            LayerScaler(init_value=layer_scaler_init_value, dimensions=features),
            alias=[""],
        )
        self.add_module("StochasticDepth", DropPath(drop_p))
        self.add_module("Residual", blocks.Add(), in_branch=[0, 1])


class DownSample(network.ModuleArgsDict):

    def __init__(self, in_features: int, out_features: int, dim: int):
        super().__init__()
        self.add_module(
            "LayerNorm",
            LayerNorm(in_features, eps=1e-6, data_format="channels_first"),
            alias=["0"],
        )
        self.add_module(
            "Conv",
            blocks.get_torch_module("Conv", dim)(in_features, out_features, kernel_size=2, stride=2),
            alias=["1"],
        )


class ConvNexStage(network.ModuleArgsDict):

    def __init__(self, features: int, depth: int, drop_p: list[float], dim: int):
        super().__init__()
        for i in range(depth):
            self.add_module(
                f"BottleNeckBlock_{i}",
                BottleNeckBlock(
                    features=features,
                    drop_p=drop_p[i],
                    layer_scaler_init_value=1e-6,
                    dim=dim,
                ),
                alias=[f"{i}"],
            )


class ConvNextStem(network.ModuleArgsDict):

    def __init__(self, in_features: int, out_features: int, dim: int):
        super().__init__()
        self.add_module(
            "Conv",
            blocks.get_torch_module("Conv", dim)(in_features, out_features, kernel_size=4, stride=4),
            alias=["0"],
        )
        self.add_module(
            "LayerNorm",
            LayerNorm(out_features, eps=1e-6, data_format="channels_first"),
            alias=["1"],
        )


class ConvNextEncoder(network.ModuleArgsDict):

    def __init__(
        self,
        in_channels: int,
        depths: list[int],
        widths: list[int],
        drop_p: float,
        dim: int,
    ):
        super().__init__()
        self.add_module(
            "ConvNextStem",
            ConvNextStem(in_channels, widths[0], dim=dim),
            alias=["downsample_layers.0"],
        )

        drop_probs = [x.item() for x in torch.linspace(0, drop_p, sum(depths))]
        self.add_module(
            "ConvNexStage_0",
            ConvNexStage(
                features=widths[0],
                depth=depths[0],
                drop_p=drop_probs[: depths[0]],
                dim=dim,
            ),
            alias=["stages.0"],
        )

        for i, (in_features, out_features) in enumerate(list(zip(widths[:], widths[1:]))):
            self.add_module(
                f"DownSample_{i + 1}",
                DownSample(in_features=in_features, out_features=out_features, dim=dim),
                alias=[f"downsample_layers.{i + 1}"],
            )
            self.add_module(
                f"ConvNexStage_{i + 1}",
                ConvNexStage(
                    features=out_features,
                    depth=depths[i + 1],
                    drop_p=drop_probs[sum(depths[: i + 1]) : sum(depths[: i + 2])],
                    dim=dim,
                ),
                alias=[f"stages.{i + 1}"],
            )


class Head(network.ModuleArgsDict):

    def __init__(self, in_features: int, num_classes: list[int], dim: int) -> None:
        super().__init__()
        self.add_module(
            "AdaptiveAvgPool",
            blocks.get_torch_module("AdaptiveAvgPool", dim)(tuple([1] * dim)),
        )
        self.add_module("Flatten", torch.nn.Flatten(1))
        self.add_module("LayerNorm", torch.nn.LayerNorm(in_features, eps=1e-6), alias=["norm"])

        for i, nb_classe in enumerate(num_classes):
            self.add_module(
                f"Linear_{i}",
                torch.nn.Linear(in_features, nb_classe),
                pretrained=False,
                alias=["head"],
                out_branch=[i + 1],
            )
            self.add_module(
                f"Unsqueeze_{i}",
                blocks.Unsqueeze(2),
                in_branch=[i + 1],
                out_branch=[-1],
            )


class ConvNeXt(network.Network):

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch = ModelPatch(),
        dim: int = 3,
        in_channels: int = 1,
        depths: list[int] = [3, 3, 27, 3],
        widths: list[int] = [128, 256, 512, 1024],
        drop_p: float = 0.1,
        num_classes: list[int] = [4, 7],
    ):

        super().__init__(
            in_channels=in_channels,
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            dim=dim,
            patch=patch,
            init_type="trunc_normal",
            init_gain=0.02,
        )
        self.add_module(
            "ConvNextEncoder",
            ConvNextEncoder(
                in_channels=in_channels,
                depths=depths,
                widths=widths,
                drop_p=drop_p,
                dim=dim,
            ),
        )
        self.add_module("Head", Head(in_features=widths[-1], num_classes=num_classes, dim=dim))



import torch

from konfai.network import blocks, network


class ConvBlock(torch.nn.Module):

    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.Conv_0 = torch.nn.Conv3d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
        )
        self.Norm_0 = torch.nn.InstanceNorm3d(num_features=out_channels, affine=True)
        self.Activation_0 = torch.nn.LeakyReLU(negative_slope=0.01)
        self.Conv_1 = torch.nn.Conv3d(
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
        )
        self.Norm_1 = torch.nn.InstanceNorm3d(num_features=out_channels, affine=True)
        self.Activation_1 = torch.nn.LeakyReLU(negative_slope=0.01)

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        output = self.Conv_0(tensor)
        output = self.Norm_0(output)
        output = self.Activation_0(output)
        output = self.Conv_1(output)
        output = self.Norm_1(output)
        output = self.Activation_1(output)
        return output


class UnetCPP1Layers(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()
        self.DownConvBlock_0 = ConvBlock(in_channels=1, out_channels=32)

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return self.DownConvBlock_0(tensor)


class Adaptation(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()
        self.Encoder_1 = UnetCPP1Layers()
        self.ToFeatures = blocks.ToFeatures(3)
        self.FCT_1 = torch.nn.Linear(32, 32, bias=True)

    def forward(
        self, a: torch.Tensor, b: torch.Tensor, c: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        self.Encoder_1.requires_grad_(False)
        self.FCT_1.requires_grad_(True)
        return (
            self.FCT_1(self.ToFeatures(self.Encoder_1(a))),
            self.FCT_1(self.ToFeatures(self.Encoder_1(b))),
            self.FCT_1(self.ToFeatures(self.Encoder_1(c))),
        )


class Representation(network.Network):

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        dim: int = 3,
    ):
        super().__init__(
            in_channels=1,
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            dim=dim,
            init_type="kaiming",
        )
        self.add_module("Model", Adaptation(), in_branch=[0, 1, 2])



import torch

from konfai.data.patching import ModelPatch
from konfai.network import blocks, network


class NestedUNet(network.Network):

    class NestedUNetBlock(network.ModuleArgsDict):

        def __init__(
            self,
            channels: list[int],
            nb_conv_per_stage: int,
            block_config: blocks.BlockConfig,
            downsample_mode: blocks.DownsampleMode,
            upsample_mode: blocks.UpsampleMode,
            attention: bool,
            block: type,
            dim: int,
            i: int = 0,
        ) -> None:
            super().__init__()
            if i > 0:
                self.add_module(
                    downsample_mode.name,
                    blocks.downsample(
                        in_channels=channels[0],
                        out_channels=channels[1],
                        downsample_mode=downsample_mode,
                        dim=dim,
                    ),
                )

            self.add_module(
                f"X_{i}_{0}",
                block(
                    in_channels=channels[(1 if downsample_mode == blocks.DownsampleMode.CONV_STRIDE and i > 0 else 0)],
                    out_channels=channels[1],
                    block_configs=[block_config] * nb_conv_per_stage,
                    dim=dim,
                ),
                out_branch=[f"X_{i}_{0}"],
            )
            if len(channels) > 2:
                self.add_module(
                    f"UNetBlock_{i + 1}",
                    NestedUNet.NestedUNetBlock(
                        channels[1:],
                        nb_conv_per_stage,
                        block_config,
                        downsample_mode,
                        upsample_mode,
                        attention,
                        block,
                        dim,
                        i + 1,
                    ),
                    in_branch=[f"X_{i}_{0}"],
                    out_branch=[f"X_{i + 1}_{j}" for j in range(len(channels) - 2)],
                )
                for j in range(len(channels) - 2):
                    self.add_module(
                        f"X_{i}_{j + 1}_{upsample_mode.name}",
                        blocks.upsample(
                            in_channels=channels[2],
                            out_channels=channels[1],
                            upsample_mode=upsample_mode,
                            dim=dim,
                        ),
                        in_branch=[f"X_{i + 1}_{j}"],
                        out_branch=[f"X_{i + 1}_{j}"],
                    )
                    self.add_module(
                        f"SkipConnection_{i}_{j + 1}",
                        blocks.Concat(),
                        in_branch=[f"X_{i + 1}_{j}"] + [f"X_{i}_{r}" for r in range(j + 1)],
                        out_branch=[f"X_{i}_{j + 1}"],
                    )
                    self.add_module(
                        f"X_{i}_{j + 1}",
                        block(
                            in_channels=(
                                (channels[1] * (j + 1) + channels[2])
                                if upsample_mode != blocks.UpsampleMode.CONV_TRANSPOSE
                                else channels[1] * (j + 2)
                            ),
                            out_channels=channels[1],
                            block_configs=[block_config] * nb_conv_per_stage,
                            dim=dim,
                        ),
                        in_branch=[f"X_{i}_{j + 1}"],
                        out_branch=[f"X_{i}_{j + 1}"],
                    )

    class NestedUNetHead(network.ModuleArgsDict):

        def __init__(self, in_channels: int, nb_class: int, activation: str, dim: int) -> None:
            super().__init__()
            self.add_module(
                "Conv",
                blocks.get_torch_module("Conv", dim)(
                    in_channels=in_channels,
                    out_channels=nb_class,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                ),
            )
            if activation == "Softmax":
                self.add_module("Softmax", torch.nn.Softmax(dim=1))
                self.add_module("Argmax", blocks.ArgMax(dim=1))
            elif activation == "Tanh":
                self.add_module("Tanh", torch.nn.Tanh())

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch | None = None,
        dim: int = 3,
        channels: list[int] = [1, 64, 128, 256, 512, 1024],
        nb_class: int = 2,
        block_config: blocks.BlockConfig = blocks.BlockConfig(),
        nb_conv_per_stage: int = 2,
        downsample_mode: str = "MAXPOOL",
        upsample_mode: str = "CONV_TRANSPOSE",
        attention: bool = False,
        block_type: str = "Conv",
        activation: str = "Softmax",
    ) -> None:
        super().__init__(
            in_channels=channels[0],
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            patch=patch,
            dim=dim,
        )

        self.add_module(
            "UNetBlock_0",
            NestedUNet.NestedUNetBlock(
                channels,
                nb_conv_per_stage,
                block_config,
                downsample_mode=blocks.DownsampleMode[downsample_mode],
                upsample_mode=blocks.UpsampleMode[upsample_mode],
                attention=attention,
                block=blocks.ConvBlock if block_type == "Conv" else blocks.ResBlock,
                dim=dim,
            ),
            out_branch=[f"X_0_{j + 1}" for j in range(len(channels) - 2)],
        )
        for j in range(len(channels) - 2):
            self.add_module(
                f"Head_{j}",
                NestedUNet.NestedUNetHead(
                    in_channels=channels[1],
                    nb_class=nb_class,
                    activation=activation,
                    dim=dim,
                ),
                in_branch=[f"X_0_{j + 1}"],
                out_branch=[-1],
            )


class UNetpp(network.Network):

    class ResNetEncoderLayer(network.ModuleArgsDict):

        def __init__(
            self,
            in_channel: int,
            out_channel: int,
            nb_block: int,
            dim: int,
            downsample_mode: blocks.DownsampleMode,
        ):
            super().__init__()
            for i in range(nb_block):
                if downsample_mode == blocks.DownsampleMode.MAXPOOL and i == 0:
                    self.add_module(
                        "DownSample",
                        blocks.get_torch_module("MaxPool", dim)(
                            kernel_size=3,
                            stride=2,
                            padding=1,
                            dilation=1,
                            ceil_mode=False,
                        ),
                    )
                self.add_module(
                    f"ResBlock_{i}",
                    blocks.ResBlock(
                        in_channel,
                        out_channel,
                        [
                            blocks.BlockConfig(
                                3,
                                (2 if downsample_mode == blocks.DownsampleMode.CONV_STRIDE and i == 0 else 1),
                                1,
                                False,
                                "ReLU;True",
                                blocks.NormMode.BATCH,
                            ),
                            blocks.BlockConfig(3, 1, 1, False, None, blocks.NormMode.BATCH),
                        ],
                        dim=dim,
                    ),
                )
                in_channel = out_channel

    @staticmethod
    def resnet_encoder(channels: list[int], layers: list[int], dim: int) -> list[torch.nn.Module]:
        modules = []
        modules.append(
            blocks.ConvBlock(
                channels[0],
                channels[1],
                [blocks.BlockConfig(7, 2, 3, False, "ReLU", blocks.NormMode.BATCH)],
                dim=dim,
            )
        )
        for i, (in_channel, out_channel, layer) in enumerate(zip(channels[1:], channels[2:], layers)):
            modules.append(
                UNetpp.ResNetEncoderLayer(
                    in_channel,
                    out_channel,
                    layer,
                    dim,
                    (blocks.DownsampleMode.MAXPOOL if i == 0 else blocks.DownsampleMode.CONV_STRIDE),
                )
            )
        return modules

    class UNetPPBlock(network.ModuleArgsDict):

        def __init__(
            self,
            encoder_channels: list[int],
            decoder_channels: list[int],
            encoders: list[torch.nn.Module],
            upsample_mode: blocks.UpsampleMode,
            dim: int,
            i: int = 0,
        ) -> None:
            super().__init__()
            self.add_module(f"X_{i}_{0}", encoders[0], out_branch=[f"X_{i}_{0}"])
            if len(encoder_channels) > 2:
                self.add_module(
                    f"UNetBlock_{i + 1}",
                    UNetpp.UNetPPBlock(
                        encoder_channels[1:],
                        decoder_channels[1:],
                        encoders[1:],
                        upsample_mode,
                        dim,
                        i + 1,
                    ),
                    in_branch=[f"X_{i}_{0}"],
                    out_branch=[f"X_{i + 1}_{j}" for j in range(len(encoder_channels) - 2)],
                )
                for j in range(len(encoder_channels) - 2):
                    in_channels = (
                        decoder_channels[3]
                        if j == len(encoder_channels) - 3 and len(encoder_channels) > 3
                        else encoder_channels[2]
                    )
                    out_channel = decoder_channels[2] if j == len(encoder_channels) - 3 else encoder_channels[1]
                    self.add_module(
                        f"X_{i}_{j + 1}_{upsample_mode.name}",
                        blocks.upsample(
                            in_channels=in_channels,
                            out_channels=out_channel,
                            upsample_mode=upsample_mode,
                            dim=dim,
                        ),
                        in_branch=[f"X_{i + 1}_{j}"],
                        out_branch=[f"X_{i + 1}_{j}"],
                    )
                    self.add_module(
                        f"SkipConnection_{i}_{j + 1}",
                        blocks.Concat(),
                        in_branch=[f"X_{i + 1}_{j}"] + [f"X_{i}_{r}" for r in range(j + 1)],
                        out_branch=[f"X_{i}_{j + 1}"],
                    )
                    self.add_module(
                        f"X_{i}_{j + 1}",
                        blocks.ConvBlock(
                            in_channels=encoder_channels[1] * (j + 1)
                            + (in_channels if upsample_mode != blocks.UpsampleMode.CONV_TRANSPOSE else out_channel),
                            out_channels=out_channel,
                            block_configs=[blocks.BlockConfig(3, 1, 1, False, "ReLU;True", blocks.NormMode.BATCH)] * 2,
                            dim=dim,
                        ),
                        in_branch=[f"X_{i}_{j + 1}"],
                        out_branch=[f"X_{i}_{j + 1}"],
                    )

    class UNetPPHead(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, nb_class: int, dim: int) -> None:
            super().__init__()
            self.add_module(
                "Upsample",
                blocks.upsample(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    upsample_mode=blocks.UpsampleMode.UPSAMPLE,
                    dim=dim,
                ),
            )
            self.add_module(
                "ConvBlock",
                blocks.ConvBlock(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    block_configs=[blocks.BlockConfig(3, 1, 1, False, "ReLU;True", blocks.NormMode.BATCH)] * 2,
                    dim=dim,
                ),
            )
            self.add_module(
                "Conv",
                blocks.get_torch_module("Conv", dim)(
                    in_channels=out_channels,
                    out_channels=nb_class,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                ),
            )
            if nb_class > 1:
                self.add_module("Softmax", torch.nn.Softmax(dim=1))
                self.add_module("Argmax", blocks.ArgMax(dim=1))
            else:
                self.add_module("Tanh", torch.nn.Tanh())

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch | None = None,
        encoder_channels: list[int] = [1, 64, 64, 128, 256, 512],
        decoder_channels: list[int] = [256, 128, 64, 32, 16, 1],
        layers: list[int] = [3, 4, 6, 3],
        dim: int = 2,
    ) -> None:
        super().__init__(
            in_channels=encoder_channels[0],
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            patch=patch,
            dim=dim,
        )
        self.add_module(
            "Block_0",
            UNetpp.UNetPPBlock(
                encoder_channels,
                decoder_channels[::-1],
                UNetpp.resnet_encoder(encoder_channels, layers, dim),
                blocks.UpsampleMode.UPSAMPLE,
                dim=dim,
            ),
            out_branch=[f"X_0_{j + 1}" for j in range(len(encoder_channels) - 2)],
        )
        self.add_module(
            "Head",
            UNetpp.UNetPPHead(
                in_channels=decoder_channels[-3],
                out_channels=decoder_channels[-2],
                nb_class=decoder_channels[-1],
                dim=dim,
            ),
            in_branch=[f"X_0_{len(encoder_channels) - 2}"],
            out_branch=[-1],
        )



from abc import ABC

import torch

from konfai.data.patching import ModelPatch
from konfai.network import blocks, network

"""
'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
dim = 2, in_channels = 3, depths=[2, 2, 2, 2], widths = [64, 64, 128, 256, 512],
num_classes=1000, use_bottleneck=False

'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
dim = 2, in_channels = 3, depths=[3, 4, 6, 3], widths = [64, 64, 128, 256, 512],
num_classes=1000, use_bottleneck=False

'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
dim = 2, in_channels = 3, depths=[3, 4, 6, 3], widths = [64, 256, 512, 1024, 2048],
num_classes=1000, use_bottleneck=True

'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
dim = 2, in_channels = 3, depths=[3, 4, 23, 3], widths = [64, 256, 512, 1024, 2048],
num_classes=1000, use_bottleneck=True

'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
dim = 2, in_channels = 3, depths=[3, 8, 36, 3], widths = [64, 256, 512, 1024, 2048],
num_classes=1000, use_bottleneck=True

'resnext50_32x4d': 'https://download.pytorch.org/models/resnext50_32x4d-7cdf4587.pth',
'resnext101_32x8d': 'https://download.pytorch.org/models/resnext101_32x8d-8ba56ff5.pth',
'wide_resnet50_2': 'https://download.pytorch.org/models/wide_resnet50_2-95faca4d.pth',
'wide_resnet101_2': 'https://download.pytorch.org/models/wide_resnet101_2-32ee1156.pth',
"""


class AbstractResBlock(network.ModuleArgsDict, ABC):

    def __init__(self, in_channels: int, out_channels: int, downsample: bool, dim: int):
        super().__init__()


class ResBlock(AbstractResBlock):

    def __init__(self, in_channels: int, out_channels: int, downsample: bool, dim: int):
        super().__init__(in_channels, out_channels, downsample, dim)
        if downsample:
            self.add_module(
                "Shortcut",
                blocks.ConvBlock(
                    in_channels,
                    out_channels,
                    blocks.BlockConfig(
                        kernel_size=1,
                        stride=2,
                        padding=0,
                        bias=False,
                        activation="None",
                        norm_mode=blocks.NormMode.BATCH.name,
                    ),
                    dim=dim,
                    alias=[["0"], ["1"], []],
                ),
                in_branch=[1],
                out_branch=[1],
                alias=["downsample"],
            )

        self.add_module(
            "ConvBlock_0",
            blocks.ConvBlock(
                in_channels,
                out_channels,
                blocks.BlockConfig(
                    kernel_size=3,
                    stride=2 if downsample else 1,
                    padding=1,
                    bias=False,
                    activation="ReLU",
                    norm_mode=blocks.NormMode.BATCH.name,
                ),
                dim=dim,
                alias=[["conv1"], ["bn1"], []],
            ),
        )
        self.add_module(
            "ConvBlock_1",
            blocks.ConvBlock(
                out_channels,
                out_channels,
                blocks.BlockConfig(
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=False,
                    activation="None",
                    norm_mode=blocks.NormMode.BATCH.name,
                ),
                dim=dim,
                alias=[["conv2"], ["bn2"], []],
            ),
        )
        self.add_module("Residual", blocks.Add(), in_branch=[0, 1])
        self.add_module("ReLU", torch.nn.ReLU())


class ResBottleneckBlock(AbstractResBlock):
    def __init__(self, in_channels: int, out_channels: int, downsample: bool, dim: int):
        super().__init__(in_channels, out_channels, downsample, dim)
        self.add_module(
            "ConvBlock_0",
            blocks.ConvBlock(
                in_channels,
                out_channels // 4,
                blocks.BlockConfig(
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    bias=False,
                    activation="ReLU",
                    norm_mode=blocks.NormMode.BATCH.name,
                ),
                dim=dim,
                alias=[["conv1"], ["bn1"], []],
            ),
        )
        self.add_module(
            "ConvBlock_1",
            blocks.ConvBlock(
                out_channels // 4,
                out_channels // 4,
                blocks.BlockConfig(
                    kernel_size=3,
                    stride=2 if downsample else 1,
                    padding=1,
                    bias=False,
                    activation="ReLU",
                    norm_mode=blocks.NormMode.BATCH.name,
                ),
                dim=dim,
                alias=[["conv2"], ["bn2"], []],
            ),
        )
        self.add_module(
            "ConvBlock_2",
            blocks.ConvBlock(
                out_channels // 4,
                out_channels,
                blocks.BlockConfig(
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    bias=False,
                    activation="ReLU",
                    norm_mode=blocks.NormMode.BATCH.name,
                ),
                dim=dim,
                alias=[["conv3"], ["bn3"], []],
            ),
        )

        if downsample or in_channels != out_channels:
            self.add_module(
                "Shortcut",
                blocks.ConvBlock(
                    in_channels,
                    out_channels,
                    blocks.BlockConfig(
                        kernel_size=1,
                        stride=2 if downsample else 1,
                        padding=0,
                        bias=False,
                        activation="None",
                        norm_mode=blocks.NormMode.BATCH.name,
                    ),
                    dim=dim,
                    alias=[["0"], ["1"], []],
                ),
                in_branch=[1],
                out_branch=[1],
                alias=["downsample"],
            )

        self.add_module("Residual", blocks.Add(), in_branch=[0, 1])
        self.add_module("ReLU", torch.nn.ReLU())


class ResNetStage(network.ModuleArgsDict):

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        depth: int,
        block: type[AbstractResBlock],
        downsample: bool,
        dim: int,
    ):
        super().__init__()
        self.add_module(
            "BottleNeckBlock_0",
            block(
                in_channels=in_channels,
                out_channels=out_channels,
                downsample=downsample,
                dim=dim,
            ),
            alias=["0"],
        )
        for i in range(1, depth):
            self.add_module(
                f"BottleNeckBlock_{i}",
                block(
                    in_channels=out_channels,
                    out_channels=out_channels,
                    downsample=False,
                    dim=dim,
                ),
                alias=[f"{i}"],
            )


class ResNetStem(network.ModuleArgsDict):

    def __init__(self, in_channels: int, out_features: int, dim: int):
        super().__init__()
        self.add_module(
            "ConvBlock",
            blocks.ConvBlock(
                in_channels,
                out_features,
                blocks.BlockConfig(
                    kernel_size=7,
                    stride=2,
                    padding=3,
                    bias=False,
                    activation="ReLU",
                    norm_mode=blocks.NormMode.BATCH.name,
                ),
                dim=dim,
                alias=[["conv1"], ["bn1"], []],
            ),
        )
        self.add_module(
            "MaxPool",
            blocks.get_torch_module("MaxPool", dim)(kernel_size=3, stride=2, padding=1),
        )


class ResNetEncoder(network.ModuleArgsDict):

    def __init__(
        self,
        in_channels: int,
        depths: list[int],
        widths: list[int],
        use_bottleneck: bool,
        dim: int,
    ):
        super().__init__()
        self.add_module("ResNetStem", ResNetStem(in_channels, widths[0], dim=dim))

        for i, (in_channels, out_channels, depth) in enumerate(list(zip(widths[:], widths[1:], depths))):
            self.add_module(
                f"ResNetStage_{i}",
                ResNetStage(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    depth=depth,
                    block=ResBottleneckBlock if use_bottleneck else ResBlock,
                    downsample=i != 0,
                    dim=dim,
                ),
                alias=[f"layer{i + 1}"],
            )


class Head(network.ModuleArgsDict):

    def __init__(self, in_features: int, num_classes: int, dim: int):
        super().__init__()
        self.add_module("AdaptiveAvgPool", blocks.get_torch_module("AdaptiveAvgPool", dim)(1))
        self.add_module("Flatten", torch.nn.Flatten(1))
        self.add_module(
            "Linear",
            torch.nn.Linear(in_features, num_classes),
            pretrained=False,
            alias=["fc"],
        )
        self.add_module("Unsqueeze", blocks.Unsqueeze(2))


class ResNet(network.Network):

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch = ModelPatch(),
        dim: int = 3,
        in_channels: int = 1,
        depths: list[int] = [2, 2, 2, 2],
        widths: list[int] = [64, 64, 128, 256, 512],
        num_classes: int = 10,
        use_bottleneck=False,
    ):
        super().__init__(
            in_channels=in_channels,
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            dim=dim,
            patch=patch,
            init_type="trunc_normal",
            init_gain=0.02,
        )
        self.add_module(
            "ResNetEncoder",
            ResNetEncoder(
                in_channels=in_channels,
                depths=depths,
                widths=widths,
                use_bottleneck=use_bottleneck,
                dim=dim,
            ),
        )
        self.add_module("Head", Head(in_features=widths[-1], num_classes=num_classes, dim=dim))



import datetime
import os


def models_directory() -> str:
    return _get_env("KONFAI_MODELS_DIRECTORY")


def checkpoints_directory() -> str:
    return _get_env("KONFAI_CHECKPOINTS_DIRECTORY")


def path_to_models() -> str:
    return _get_env("KONFAI_MODEL")


def predictions_directory() -> str:
    return _get_env("KONFAI_PREDICTIONS_DIRECTORY")


def evaluations_directory() -> str:
    return _get_env("KONFAI_EVALUATIONS_DIRECTORY")


def statistics_directory() -> str:
    return _get_env("KONFAI_STATISTICS_DIRECTORY")


def setups_directory() -> str:
    return _get_env("KONFAI_SETUPS_DIRECTORY")


def config_file() -> str:
    return _get_env("KONFAI_config_file")


def konfai_state() -> str:
    return _get_env("KONFAI_STATE")


def konfai_root() -> str:
    return _get_env("konfai_root")


def cuda_visible_devices() -> str:
    return _get_env("CUDA_VISIBLE_DEVICES")


def konfai_nb_cores() -> str:
    return _get_env("KONFAI_NB_CORES")


def current_date() -> str:
    return datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")


def _get_env(var: str) -> str:
    value = os.environ.get(var)
    if value is None:
        raise RuntimeError(f"Environment variable '{var}' is not set.")
    return value
